"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;

  function optional(path, replacementPath) {
    var loose = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
    var scope = path.scope;

    var optionals = [];
    var nil = scope.buildUndefinedNode();

    var objectPath = path;
    while (objectPath.isMemberExpression() || objectPath.isCallExpression() || objectPath.isNewExpression()) {
      var _objectPath = objectPath,
          node = _objectPath.node;

      if (node.optional) {
        optionals.push(node);
      }

      if (objectPath.isMemberExpression()) {
        objectPath = objectPath.get("object");
      } else {
        objectPath = objectPath.get("callee");
      }
    }

    for (var i = optionals.length - 1; i >= 0; i--) {
      var node = optionals[i];
      node.optional = false;

      var isCall = t.isCallExpression(node);
      var replaceKey = isCall || t.isNewExpression(node) ? "callee" : "object";
      var chain = node[replaceKey];

      var ref = void 0;
      var check = void 0;
      if (loose && isCall) {
        // If we are using a loose transform (avoiding a Function#call) and we are at the call,
        // we can avoid a needless memoize.
        check = ref = chain;
      } else {
        ref = scope.maybeGenerateMemoised(chain);
        if (ref) {
          check = t.assignmentExpression("=", ref, chain);
          node[replaceKey] = ref;
        } else {
          check = ref = chain;
        }
      }

      // Ensure call expressions have the proper `this`
      // `foo.bar()` has context `foo`.
      if (isCall && t.isMemberExpression(chain)) {
        if (loose) {
          // To avoid a Function#call, we can instead re-grab the property from the context object.
          // `a.?b.?()` translates roughly to `_a.b != null && _a.b()`
          node.callee = chain;
        } else {
          // Otherwise, we need to memoize the context object, and change the call into a Function#call.
          // `a.?b.?()` translates roughly to `(_b = _a.b) != null && _b.call(_a)`
          var object = chain.object;

          var context = scope.maybeGenerateMemoised(object);
          if (context) {
            chain.object = t.assignmentExpression("=", context, object);
          } else {
            context = object;
          }

          node.arguments.unshift(context);
          node.callee = t.memberExpression(node.callee, t.identifier("call"));
        }
      }

      replacementPath.replaceWith(t.conditionalExpression(t.binaryExpression("==", check, t.nullLiteral()), nil, replacementPath.node));

      replacementPath = replacementPath.get("alternate");
    }
  }

  function findReplacementPath(path) {
    return path.find(function (path) {
      var parentPath = path.parentPath;


      if (path.key == "left" && parentPath.isAssignmentExpression()) {
        return false;
      }
      if (path.key == "object" && parentPath.isMemberExpression()) {
        return false;
      }
      if (path.key == "callee" && (parentPath.isCallExpression() || parentPath.isNewExpression())) {
        return false;
      }
      if (path.key == "argument" && parentPath.isUpdateExpression()) {
        return false;
      }
      if (path.key == "argument" && parentPath.isUnaryExpression({ operator: "delete" })) {
        return false;
      }

      return true;
    });
  }

  return {
    inherits: _babelPluginSyntaxOptionalChaining2.default,

    visitor: {
      "MemberExpression|NewExpression|CallExpression": function MemberExpressionNewExpressionCallExpression(path) {
        if (!path.node.optional) {
          return;
        }

        optional(path, findReplacementPath(path), this.opts.loose);
      }
    }
  };
};

var _babelPluginSyntaxOptionalChaining = require("babel-plugin-syntax-optional-chaining");

var _babelPluginSyntaxOptionalChaining2 = _interopRequireDefault(_babelPluginSyntaxOptionalChaining);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }